package cc.fascinated.bat.service;

import cc.fascinated.bat.common.DateUtils;
import cc.fascinated.bat.common.WebRequest;
import cc.fascinated.bat.exception.BadRequestException;
import cc.fascinated.bat.exception.ResourceNotFoundException;
import cc.fascinated.bat.model.beatsaber.scoresaber.ScoreSaberAccountToken;
import cc.fascinated.bat.model.beatsaber.scoresaber.ScoreSaberPageMetadataToken;
import cc.fascinated.bat.model.beatsaber.scoresaber.ScoreSaberPlayerScoreToken;
import cc.fascinated.bat.model.beatsaber.scoresaber.ScoreSaberScoresPageToken;
import cc.fascinated.bat.model.user.profiles.ScoreSaberProfile;
import lombok.extern.log4j.Log4j2;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

@Service @Log4j2(topic = "ScoreSaber Service")
public class ScoreSaberService {
    private static final String SCORESABER_API = "https://scoresaber.com/api/";
    private static final String GET_PLAYER_ENDPOINT = SCORESABER_API + "player/%s/full";
    private static final String GET_PLAYER_SCORES_ENDPOINT = SCORESABER_API + "player/%s/scores?limit=100&sort=%s&page=%s&withMetadata=true";

    /**
     * Gets the account from the ScoreSaber API.
     *
     * @param id The id of the account.
     * @return The account.
     * @throws ResourceNotFoundException If the account is not found.
     * @throws cc.fascinated.bat.exception.RateLimitException If the ScoreSaber rate limit is reached.
     */
    public ScoreSaberAccountToken getAccount(String id) {
        ScoreSaberAccountToken account = WebRequest.getAsEntity(String.format(GET_PLAYER_ENDPOINT, id), ScoreSaberAccountToken.class);
        if (account == null) { // Check if the account doesn't exist.
            log.info("Account with id '{}' not found.", id);
            throw new ResourceNotFoundException("Account with id '%s' not found.".formatted(id));
        }
        if (account.isBanned()) {
            throw new BadRequestException("Account with id '%s' is banned.".formatted(id));
        }
        if (account.isInactive()) {
            throw new BadRequestException("Account with id '%s' is inactive.".formatted(id));
        }
        return account;
    }

    /**
     * Gets the scores for the account.
     *
     * @param profile The profile.
     * @param page The page to get the scores from.
     * @return The scores.
     */
    public ScoreSaberScoresPageToken getPageScores(ScoreSaberProfile profile, int page) {
        log.info("Fetching scores for account '{}' from page {}.", profile.getId(), page);
        ScoreSaberScoresPageToken pageToken = WebRequest.getAsEntity(String.format(GET_PLAYER_SCORES_ENDPOINT, profile.getId(), "recent", page), ScoreSaberScoresPageToken.class);
        if (pageToken == null) { // Check if the page doesn't exist.
            return null;
        }
        // Sort the scores by newest time set.
        pageToken.setPlayerScores(Arrays.stream(pageToken.getPlayerScores())
                .sorted((a, b) -> DateUtils.getDateFromString(b.getScore().getTimeSet()).compareTo(DateUtils.getDateFromString(a.getScore().getTimeSet())))
                .toArray(ScoreSaberPlayerScoreToken[]::new));
        return pageToken;
    }

    /**
     * Gets the scores for the account.
     *
     * @param profile The profile.
     * @return The scores.
     */
    public List<ScoreSaberScoresPageToken> getScores(ScoreSaberProfile profile) {
        List<ScoreSaberScoresPageToken> scores = new ArrayList<>(List.of(getPageScores(profile, 1)));
        ScoreSaberPageMetadataToken metadata = scores.get(0).getMetadata();
        int totalPages = (int) Math.ceil((double) metadata.getTotal() / metadata.getItemsPerPage());
        log.info("Fetching {} pages of scores for account '{}'.", totalPages, profile.getId());
        for (int i = 2; i <= totalPages; i++) {
            scores.add(getPageScores(profile, i));
        }

        return scores;
    }
}
